/*mkmed incdir:guile/include */
/*mkmed incdir:openldap/include */
/*mkmed :arg:-fPIC */

/* struct timeval */
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>

#include <libguile.h>
#include <ldap.h>

#ifndef LDAP_OPT_RESULT_CODE
#define LDAP_OPT_RESULT_CODE LDAP_OPT_ERROR_NUMBER
#endif

/* snarfing macros */

#define SCMLDAP_SYMBOL(name) \
  SCM_SYMBOL(scmldap_##name, "LDAP_" #name)
#define SCMLDAP_VAR(name, value) \
  SCM_VARIABLE_INIT(scmldap_##name, "LDAP_" #name, value)
#define SCMLDAP_VAR_INT(name) \
  SCMLDAP_VAR(name, scm_int2num(LDAP_##name))
#define SCMLDAP_SMOB(c_name, scheme_name) \
  SCM_SNARF_HERE(static scm_t_bits scmldap_##c_name##_tag) \
  SCM_SNARF_INIT(scmldap_##c_name##_tag= \
                 scm_make_smob_type((char*)(scheme_name), 0);) 
#define SCMLDAP_SMOB_FUNC(c_name, func) \
  SCM_SNARF_HERE(static size_t scmldap_##c_name##_##func(SCM smob)) \
  SCM_SNARF_INIT(scm_set_smob_##func \
                 (scmldap_##c_name##_tag, scmldap_##c_name##_##func);)

/* symbols */

SCM_SYMBOL(scmldap_error_key, "ldap:error");
SCMLDAP_SYMBOL(SASL_SIMPLE);
SCMLDAP_SYMBOL(OPT_ON);
SCMLDAP_SYMBOL(OPT_OFF);

/* variables */

SCMLDAP_VAR_INT(SCOPE_BASE);
SCMLDAP_VAR_INT(SCOPE_ONELEVEL);
SCMLDAP_VAR_INT(SCOPE_SUBTREE);
SCMLDAP_VAR_INT(MSG_ONE);
SCMLDAP_VAR_INT(MSG_ALL);
SCMLDAP_VAR_INT(MSG_RECEIVED);
SCMLDAP_VAR_INT(OPT_PROTOCOL_VERSION);
SCMLDAP_VAR_INT(VERSION2);
SCMLDAP_VAR_INT(VERSION3);

/* utility functions */

static void scmldap_throw(char const* subr, char const* errstring,
                          char const* errtext, int errcode) {
  char const* format;
  SCM args;
  if (errtext==NULL) {
    format="~A";
    args=scm_list_1(scm_makfrom0str(errstring));
  } else {
    format="~A: ~A";
    args=scm_list_2(scm_makfrom0str(errstring),
                    scm_makfrom0str(errtext));
  }
  scm_error(scmldap_error_key, subr, format, args,
            (errcode==-1
             ? SCM_EOL
             : scm_cons(scm_int2num(errcode), SCM_EOL)));
}

static void scmldap_error(char const* subr, int errcode, char const* errtext) {
  if (errcode!=LDAP_SUCCESS) {
    char const* errstring=(errcode==-1
                           ? "unknown error"
                           : ldap_err2string(errcode));
    scmldap_throw(subr, errstring, errtext, errcode);
  }
}
#define SCMLDAP_ERROR(errcode) scmldap_error(FUNC_NAME, (errcode), NULL)

static void scmldap_session_error(char const* subr, LDAP* ld) {
  int errcode;
  char const* errtext;
  if (ldap_get_option(ld, LDAP_OPT_RESULT_CODE,  &errcode)!=0)
    scmldap_throw(subr, "unknown error", NULL, -1);
  else {
    if (ldap_get_option(ld, LDAP_OPT_ERROR_STRING, &errtext)!=0)
      errtext=NULL;
    scmldap_error(subr, errcode, errtext);
  }
}
#define SCMLDAP_SESSION_ERROR(ld) scmldap_session_error(FUNC_NAME, (ld))

/*
LDAP_OPT_MATCHED_DN (0x33) char ** 
The matched DN value returned with the most recent LDAP error that occurred
for this session.
*/

static void scmldap_clear_error(char const* subr, LDAP* ld) {
  int errcode=LDAP_SUCCESS;
  if (ldap_set_option(ld, LDAP_OPT_RESULT_CODE, &errcode)!=0)
    scmldap_session_error(subr, ld);
}
#define SCMLDAP_CLEAR_ERROR(ld) scmldap_clear_error(FUNC_NAME, (ld))

/* smobs */

SCMLDAP_SMOB(session, "ldap-session");
#define SCMLDAP_SESSIONP(obj) SCM_SMOB_PREDICATE(scmldap_session_tag, obj)
#define SCMLDAP_VALIDATE_SESSION(pos, obj) \
  SCM_ASSERT_TYPE(SCMLDAP_SESSIONP(obj), obj, pos, FUNC_NAME, "LDAP_SESSIONP")
#define SCMLDAP_NEW_SESSION(obj, ld) \
  SCM_NEWSMOB((obj), scmldap_session_tag, (scm_t_bits)(ld))
#define SCMLDAP_SESSION_DATA(obj) ((LDAP*)SCM_SMOB_DATA(obj))
#define SCMLDAP_SESSION_SET_DATA(obj, x) SCM_SET_SMOB_DATA(obj, (scm_t_bits)x)

SCMLDAP_SMOB_FUNC(session, free)
{
  LDAP* ld=SCMLDAP_SESSION_DATA(smob);
  if (ld!=NULL)
    scmldap_error("c:scmldap_session_free", ldap_unbind_ext_s(ld, NULL, NULL),
                  NULL);
  return 0;
}

SCMLDAP_SMOB(msgid, "ldap-msgid");
#define SCMLDAP_MSGIDP(obj) SCM_SMOB_PREDICATE(scmldap_msgid_tag, obj)
#define SCMLDAP_VALIDATE_MSGID(pos, obj) \
  SCM_ASSERT_TYPE(SCMLDAP_MSGIDP(obj), obj, pos, FUNC_NAME, "LDAP_MSGIDP")
#define SCMLDAP_NEW_MSGID(obj, ld, msgid) \
  SCM_NEWSMOB2((obj), scmldap_msgid_tag, SCM_UNPACK(ld), (scm_t_bits)(msgid))
#define SCMLDAP_MSGID_SESS(obj) SCM_PACK(SCM_CELL_WORD_1((obj)))
#define SCMLDAP_MSGID_ID(  obj)    ((int)SCM_CELL_WORD_2((obj)))
#define SCMLDAP_MSGID_SET_SESS(obj, ld) \
  SCM_SET_CELL_WORD_1((obj), SCM_UNPACK(ld))
#define SCMLDAP_MSGID_SET_ID(  obj, id) \
  SCM_SET_CELL_WORD_2((obj), (scm_t_bits)(id))
SCM_SNARF_INIT(scm_set_smob_mark(scmldap_msgid_tag, scm_markcdr);)

SCMLDAP_SMOB(message, "ldap-message");
#define SCMLDAP_MESSAGEP(obj) SCM_SMOB_PREDICATE(scmldap_message_tag, obj)
#define SCMLDAP_VALIDATE_MESSAGE(pos, obj) \
  SCM_ASSERT_TYPE(SCMLDAP_MESSAGEP(obj), obj, pos, FUNC_NAME, "LDAP_MESSAGEP")
#define SCMLDAP_NEW_MESSAGE(obj, ld, message) \
  SCM_NEWSMOB2((obj), scmldap_message_tag, SCM_UNPACK(ld), \
               (scm_t_bits)(message))
#define SCMLDAP_MESSAGE_SESS(obj)       SCM_PACK(SCM_CELL_WORD_1((obj)))
#define SCMLDAP_MESSAGE_MSG( obj) ((LDAPMessage*)SCM_CELL_WORD_2((obj)))
#define SCMLDAP_MESSAGE_SET_SESS(obj, ld) \
  SCM_SET_CELL_WORD_1((obj), SCM_UNPACK(ld))
#define SCMLDAP_MESSAGE_SET_MSG( obj, msg) \
  SCM_SET_CELL_WORD_2((obj), (scm_t_bits)(msg))
SCM_SNARF_INIT(scm_set_smob_mark(scmldap_message_tag, scm_markcdr);)

SCMLDAP_SMOB_FUNC(message, free)
{
  ldap_msgfree(SCMLDAP_MESSAGE_MSG(smob));
  return 0;
}

SCMLDAP_SMOB(entry, "ldap-entry");
#define SCMLDAP_ENTRYP(obj) SCM_SMOB_PREDICATE(scmldap_entry_tag, obj)
#define SCMLDAP_VALIDATE_ENTRY(pos, obj) \
  SCM_ASSERT_TYPE(SCMLDAP_ENTRYP(obj), obj, pos, FUNC_NAME, "LDAP_ENTRYP")
#define SCMLDAP_NEW_ENTRY(obj, message, entry) \
  SCM_NEWSMOB2((obj), scmldap_entry_tag, SCM_UNPACK(message), \
               (scm_t_bits)(entry))
#define SCMLDAP_ENTRY_MSG(  obj)       SCM_PACK(SCM_CELL_WORD_1((obj)))
#define SCMLDAP_ENTRY_ENTRY(obj) ((LDAPMessage*)SCM_CELL_WORD_2((obj)))
#define SCMLDAP_ENTRY_SET_MSG(  obj, message) \
  SCM_SET_CELL_WORD_1((obj), SCM_UNPACK(message))
#define SCMLDAP_ENTRY_SET_ENTRY(obj, entry) \
  SCM_SET_CELL_WORD_2((obj), (scm_t_bits)(entry))
SCM_SNARF_INIT(scm_set_smob_mark(scmldap_entry_tag, scm_markcdr);)

/*
LDAPControl    - server and client controls,
LDAPAPIInfo    - provide data about the LDAP implementation to caller,
struct ldapmod - data with modifications sent by caller to LDAP.
*/

/* subrs */

#define FUNC_NAME s_scmldap_session_p
SCM_DEFINE(scmldap_session_p, "ldap:session?", 1, 0, 0, (SCM s_obj), "\
Return #t if OBJ is an LDAP session object.")
{ return SCM_BOOL(SCMLDAP_SESSIONP(s_obj)); }
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_msgid_p
SCM_DEFINE(scmldap_msgid_p, "ldap:msgid?", 1, 0, 0, (SCM s_obj), "\
Return #t if OBJ is an LDAP msgid object.")
{ return SCM_BOOL(SCMLDAP_MSGIDP(s_obj)); }
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_message_p
SCM_DEFINE(scmldap_message_p, "ldap:message?", 1, 0, 0, (SCM s_obj), "\
Return #t if OBJ is an LDAP message object.")
{ return SCM_BOOL(SCMLDAP_MESSAGEP(s_obj)); }
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_entry_p
SCM_DEFINE(scmldap_entry_p, "ldap:entry?", 1, 0, 0, (SCM s_obj), "\
Return #t if OBJ is an LDAP entry object.")
{ return SCM_BOOL(SCMLDAP_ENTRYP(s_obj)); }
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_initialize
SCM_DEFINE(scmldap_initialize, "ldap-initialize", 1, 0, 0, (SCM s_url), "\
Return a new LDAP session object referring to URL.")
{
  SCM s_ld;
  LDAP* ld;
  SCM_VALIDATE_STRING(SCM_ARG1, s_url);
  SCM_STRING_COERCE_0TERMINATION_X(s_url);
  SCMLDAP_ERROR(ldap_initialize(&ld, SCM_STRING_CHARS(s_url)));
  SCMLDAP_NEW_SESSION(s_ld, ld);
  return s_ld;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_init
SCM_DEFINE(scmldap_init, "ldap-init", 1, 1, 0, (SCM s_host, SCM s_port), "\
Return a new LDAP session object referring to HOST and PORT.\n\
PORT defaults to LDAP_PORT (389).")
{
  SCM s_ld;
  LDAP* ld;
  int port;
  SCM_VALIDATE_STRING(SCM_ARG1, s_host);
  if (SCM_UNBNDP(s_port) || SCM_FALSEP(s_port)) port=LDAP_PORT;
  else {
    SCM_VALIDATE_INUM(SCM_ARG2, s_port);
    port=SCM_INUM(s_port);
  }
  SCM_STRING_COERCE_0TERMINATION_X(s_host);
  ld=ldap_init(SCM_STRING_CHARS(s_host), port);
  if (ld==NULL) SCM_SYSERROR;
  SCMLDAP_NEW_SESSION(s_ld, ld);
  return s_ld;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_set_option
SCM_DEFINE(scmldap_set_option, "ldap-set-option", 3, 0, 0,
           (SCM s_ld, SCM s_option, SCM s_value), "Set LD's OPTION to VALUE.")
{
  LDAP* ld;
  int option;
  union {
    int v_int;
  } value;
  void const* valuep;
  if (SCM_FALSEP(s_ld)) ld=NULL;
  else {
    SCMLDAP_VALIDATE_SESSION(SCM_ARG1, s_ld);
    ld=SCMLDAP_SESSION_DATA(s_ld);
  }
  SCM_VALIDATE_INT_COPY(SCM_ARG2, s_option, option);
  switch (option) {
    case LDAP_OPT_PROTOCOL_VERSION:
      valuep=&value.v_int;
      SCM_VALIDATE_INT_COPY(SCM_ARG3, s_value, value.v_int);
      break;
    default:
      SCM_OUT_OF_RANGE(SCM_ARG2, s_option);
  }
  if (ldap_set_option(ld, option, valuep)!=0)
    SCMLDAP_SESSION_ERROR(ld);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

/* int ldap_start_tls_s(ld, serverctrls, clientctrls) */ 

#define FUNC_NAME s_scmldap_sasl_bind
SCM_DEFINE(scmldap_sasl_bind, "ldap-sasl-bind", 6, 0, 0,
           (SCM s_ld, SCM s_dn, SCM s_mechanism, SCM s_cred, SCM s_serverctrls,
            SCM s_clientctrls), "\
Perform an asynchronous SASL bind to LD as DN via MECHANISM, authenicated by\n\
CRED, under the control of SERVERCTRLS and CLIENTCTRLS.")
{
  char const* mechanism;
  BerValue cred, * credp;
  SCM s_msgid;
  int msgid;
  SCMLDAP_VALIDATE_SESSION(SCM_ARG1, s_ld);
  SCM_VALIDATE_STRING(SCM_ARG2, s_dn);
  if (SCM_EQ_P(s_mechanism, scmldap_SASL_SIMPLE)) mechanism=LDAP_SASL_SIMPLE;
  else {
    SCM_VALIDATE_STRING(SCM_ARG3, s_mechanism);
    SCM_STRING_COERCE_0TERMINATION_X(s_mechanism);
    mechanism=SCM_STRING_CHARS(s_mechanism);
  }
  if (SCM_FALSEP(s_cred)) credp=NULL;
  else {
    SCM_VALIDATE_STRING(SCM_ARG4, s_cred);
    cred.bv_len=SCM_STRING_LENGTH(s_cred);
    cred.bv_val=SCM_STRING_CHARS (s_cred);
    credp=&cred;
  }
  if (!SCM_EQ_P(s_serverctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "server controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  if (!SCM_EQ_P(s_clientctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "client controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  SCM_STRING_COERCE_0TERMINATION_X(s_dn);
  SCM_STRING_COERCE_0TERMINATION_X(s_mechanism);
  if (ldap_sasl_bind(SCMLDAP_SESSION_DATA(s_ld),
                     SCM_STRING_CHARS(s_dn),
                     mechanism,
                     credp,
                     NULL, /*serverctrls*/
                     NULL, /*clientctrls*/
                     &msgid)!=LDAP_SUCCESS)
    SCMLDAP_SESSION_ERROR(SCMLDAP_SESSION_DATA(s_ld));
  SCMLDAP_NEW_MSGID(s_msgid, s_ld, msgid);
  return s_msgid;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_unbind_ext_s
SCM_DEFINE(scmldap_unbind_ext_s, "ldap-unbind-ext-s", 3, 0, 0,
           (SCM s_ld, SCM s_serverctrls, SCM s_clientctrls), "\
Unbind LD if it is bound, under the control of SERVERCTRLS and CLIENTCTRLS.")
{
  int err;
  LDAP* ld;
  SCMLDAP_VALIDATE_SESSION(SCM_ARG1, s_ld);
  ld=SCMLDAP_SESSION_DATA(s_ld);
  if (!SCM_EQ_P(s_serverctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "server controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  if (!SCM_EQ_P(s_clientctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "client controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  err=(ld==NULL? LDAP_SUCCESS:
       ldap_unbind_ext_s(ld,
                         NULL, /*serverctrls*/
                         NULL  /*clientctrls*/));
  SCMLDAP_SESSION_SET_DATA(s_ld, (LDAP*)NULL);
  SCMLDAP_ERROR(err);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_search_ext
SCM_DEFINE(scmldap_search_ext, "ldap-search-ext", 10, 0, 0,
           (SCM s_ld, SCM s_base, SCM s_scope, SCM s_filter, SCM s_attrs,
            SCM s_attrsonly, SCM s_serverctrls, SCM s_clientctrls,
            SCM s_timeout, SCM s_sizelimit),
           "\
Perform an asynchronous search on LD rooted at BASE, as deep as SCOPE,\n\
according to FILTER, returning ATTRS, and their values unless ATTRSONLY,\n\
under the control of SERVERCTRLS and CLIENTCTRLS, limited by TIMEOUT and\n\
SIZELIMIT.")
{
  int scope, msgid;
  struct timeval timeout, * timeoutp;
  SCM s_msgid, s_attrs_v=SCM_BOOL_F;
  char** attrs;
  SCMLDAP_VALIDATE_SESSION(SCM_ARG1, s_ld);
  SCM_VALIDATE_STRING(SCM_ARG2, s_base);
  SCM_VALIDATE_INT_COPY(SCM_ARG3, s_scope, scope);
  SCM_VALIDATE_STRING(SCM_ARG4, s_filter);
  if (SCM_FALSEP(s_attrs)) attrs=NULL;
  else {
    long llen;
    size_t clen, csize, i;
    SCM s_attrs_tail;
    llen=scm_ilength(s_attrs);
    SCM_ASSERT_RANGE(SCM_ARG5, s_attrs, llen>=0);
    for (s_attrs_tail=s_attrs;
         !SCM_EQ_P(s_attrs_tail, SCM_EOL);
         s_attrs_tail=SCM_CDR(s_attrs_tail))
      SCM_VALIDATE_STRING(SCM_ARG5, SCM_CAR(s_attrs_tail));
    clen=llen;
    clen++;
    csize=clen*sizeof (char*);
    if (clen==0 || csize/sizeof (char*)!=clen)
      scm_num_overflow(FUNC_NAME);
    attrs=scm_must_malloc(csize, "char*");
    s_attrs_v=scm_make_vector(scm_long2num(llen), SCM_BOOL_F);
    for (i=0, s_attrs_tail=s_attrs;
         !SCM_EQ_P(s_attrs_tail, SCM_EOL);
         ++i, s_attrs_tail=SCM_CDR(s_attrs_tail)) {
      SCM s_attr=SCM_CAR(s_attrs_tail);
      SCM_STRING_COERCE_0TERMINATION_X(s_attr);
      SCM_VELTS(s_attrs_v)[i]=s_attr;
      attrs[i]=SCM_STRING_CHARS(s_attr);
    }
    attrs[llen]=NULL;
  }
  if (!SCM_EQ_P(s_serverctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "server controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  if (!SCM_EQ_P(s_clientctrls, SCM_EOL))
    scm_error(scmldap_error_key, FUNC_NAME,
              "client controls are not yet supported", 
              SCM_EOL, SCM_EOL);
  if (SCM_FALSEP(s_timeout)) timeoutp=NULL;
  else {
    SCM_VALIDATE_CONS(9, s_timeout);
    SCM_VALIDATE_INUM(9, SCM_CAR(s_timeout));
    SCM_VALIDATE_INUM(9, SCM_CDR(s_timeout));
    timeout.tv_sec =SCM_INUM(SCM_CAR(s_timeout));
    timeout.tv_usec=SCM_INUM(SCM_CDR(s_timeout));
    timeoutp=&timeout;
  }
  SCM_VALIDATE_INUM(10, s_sizelimit);
  SCM_STRING_COERCE_0TERMINATION_X(s_base);
  SCM_STRING_COERCE_0TERMINATION_X(s_filter);
  if (ldap_search_ext(SCMLDAP_SESSION_DATA(s_ld),
                      SCM_STRING_CHARS(s_base),
                      scope,
                      SCM_STRING_CHARS(s_filter),
                      attrs,
                      SCM_NFALSEP(s_attrsonly),
                      NULL, /*serverctrls*/
                      NULL, /*clientctrls*/
                      timeoutp,
                      SCM_INUM(s_sizelimit),
                      &msgid)!=LDAP_SUCCESS)
    SCMLDAP_SESSION_ERROR(SCMLDAP_SESSION_DATA(s_ld));
  scm_remember_upto_here_1(s_attrs_v);
  if (attrs!=NULL) scm_must_free(attrs);
  SCMLDAP_NEW_MSGID(s_msgid, s_ld, msgid);
  return s_msgid;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_result
SCM_DEFINE(scmldap_result, "ldap-result", 3, 0, 0,
           (SCM s_msgid, SCM s_all, SCM s_timeout),
           "Wait for the results of MSGID and return the message.")
{
  LDAP* ld;
  int msgid, all, type;
  struct timeval timeout, * timeoutp;
  LDAPMessage* res;
  SCM s_ld, s_message;
  SCMLDAP_VALIDATE_MSGID(SCM_ARG1, s_msgid);
  msgid=SCMLDAP_MSGID_ID  (s_msgid);
  s_ld =SCMLDAP_MSGID_SESS(s_msgid);
  if (SCMLDAP_MESSAGEP(s_ld)) return s_ld;
  ld=SCMLDAP_SESSION_DATA(s_ld);
  if (ld==NULL)
    scm_error(scmldap_error_key, FUNC_NAME, "LDAP session has been unbound",
              SCM_EOL, SCM_EOL);
  SCM_VALIDATE_INT_COPY(SCM_ARG2, s_all, all);
  if (SCM_FALSEP(s_timeout)) { timeoutp=NULL; }
  else {
    SCM_VALIDATE_CONS(SCM_ARG3, s_timeout);
    SCM_VALIDATE_INUM(SCM_ARG3, SCM_CAR(s_timeout));
    SCM_VALIDATE_INUM(SCM_ARG3, SCM_CDR(s_timeout));
    timeout.tv_sec =SCM_INUM(SCM_CAR(s_timeout));
    timeout.tv_usec=SCM_INUM(SCM_CDR(s_timeout));
    timeoutp=&timeout;
  }
  type=ldap_result(ld, msgid, all, timeoutp, &res);
  if (type==0) return SCM_BOOL_F;
  if (type==-1) SCMLDAP_SESSION_ERROR(ld);
  SCMLDAP_NEW_MESSAGE(s_message, s_ld, res);
  SCMLDAP_MSGID_SET_SESS(s_msgid, s_message);
  return s_message;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_first_entry
SCM_DEFINE(scmldap_first_entry, "ldap-first-entry", 1, 0, 0, (SCM s_message),
           "Return the first entry from MESSAGE.")
{
  LDAP* ld;
  LDAPMessage* message, * entry;
  SCM s_entry;
  SCMLDAP_VALIDATE_MESSAGE(SCM_ARG1, s_message);
  ld=SCMLDAP_SESSION_DATA(SCMLDAP_MESSAGE_SESS(s_message));
  message=                SCMLDAP_MESSAGE_MSG (s_message);
  SCMLDAP_CLEAR_ERROR(ld);
  entry=ldap_first_entry(ld, message);
  if (entry==NULL) {
    SCMLDAP_SESSION_ERROR(ld);
    return SCM_BOOL_F;
  }
  SCMLDAP_NEW_ENTRY(s_entry, s_message, entry);
  return s_entry;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_next_entry
SCM_DEFINE(scmldap_next_entry, "ldap-next-entry", 1, 0, 0, (SCM s_entry),
           "Return the next entry after ENTRY.")
{
  LDAP* ld;
  LDAPMessage* entry, * next;
  SCM s_message, s_next;
  SCMLDAP_VALIDATE_ENTRY(SCM_ARG1, s_entry);
  s_message=SCMLDAP_ENTRY_MSG  (s_entry);
  entry    =SCMLDAP_ENTRY_ENTRY(s_entry);
  ld=SCMLDAP_SESSION_DATA(SCMLDAP_MESSAGE_SESS(s_message));
  SCMLDAP_CLEAR_ERROR(ld);
  next=ldap_next_entry(ld, entry);
  if (next==NULL) {
    SCMLDAP_SESSION_ERROR(ld);
    return SCM_BOOL_F;
  }
  SCMLDAP_NEW_ENTRY(s_next, s_message, next);
  return s_next;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_get_dn
SCM_DEFINE(scmldap_get_dn, "ldap-get-dn", 1, 0, 0, (SCM s_entry),
           "Return the distinguished name of ENTRY.")
{
  LDAP* ld;
  LDAPMessage* entry;
  SCM s_dn;
  char* dn;
  SCMLDAP_VALIDATE_ENTRY(SCM_ARG1, s_entry);
  ld=SCMLDAP_SESSION_DATA(SCMLDAP_MESSAGE_SESS(SCMLDAP_ENTRY_MSG  (s_entry)));
  entry=                                       SCMLDAP_ENTRY_ENTRY(s_entry);
  dn=ldap_get_dn(ld, entry);
  if (dn==NULL) SCMLDAP_SESSION_ERROR(ld);
  s_dn=scm_makfrom0str(dn);
  ldap_memfree(dn);
  return s_dn;
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_get_attributes
SCM_DEFINE(scmldap_get_attributes, "ldap:get-attributes", 1, 0, 0,
           (SCM s_entry),
           "Return a list of ENTRY's attributes.")
{
  LDAP* ld;
  LDAPMessage* entry;
  SCM s_attrs, s_last, s_next;
  BerElement* element=NULL;
  char* attr;
  SCMLDAP_VALIDATE_ENTRY(SCM_ARG1, s_entry);
  ld=SCMLDAP_SESSION_DATA(SCMLDAP_MESSAGE_SESS(SCMLDAP_ENTRY_MSG  (s_entry)));
  entry=                                       SCMLDAP_ENTRY_ENTRY(s_entry);
  SCMLDAP_CLEAR_ERROR(ld);
  attr=ldap_first_attribute(ld, entry, &element);
  if (attr==NULL) {
    if (element!=NULL) ber_free(element, 0);
    SCMLDAP_SESSION_ERROR(ld);
    return SCM_EOL;
  }
  s_attrs=scm_list_1(scm_makfrom0str(attr));
  ldap_memfree(attr);
  s_last=s_attrs;
  for (;;) {
    SCMLDAP_CLEAR_ERROR(ld);
    attr=ldap_next_attribute(ld, entry, element);
    if (attr==NULL) {
      if (element!=NULL) ber_free(element, 0);
      SCMLDAP_SESSION_ERROR(ld);
      return s_attrs;
    }
    s_next=scm_list_1(scm_makfrom0str(attr));
    ldap_memfree(attr);
    scm_set_cdr_x(s_last, s_next);
    s_last=s_next;
  }
}
#undef FUNC_NAME

#define FUNC_NAME s_scmldap_get_values_len
SCM_DEFINE(scmldap_get_values_len, "ldap-get-values-len", 2, 0, 0,
           (SCM s_entry, SCM s_attr),
           "Return a list of values in ENTRY for ATTRIBUTE.")
{
  LDAP* ld;
  LDAPMessage* entry;
  char const* attr;
  BerValue** vals;
  int ct;
  SCM s_vals;
  SCMLDAP_VALIDATE_ENTRY (SCM_ARG1, s_entry);
  SCM_VALIDATE_STRING(SCM_ARG2, s_attr);
  entry=SCMLDAP_ENTRY_ENTRY(s_entry);
  ld=SCMLDAP_SESSION_DATA(SCMLDAP_MESSAGE_SESS(SCMLDAP_ENTRY_MSG(s_entry)));
  SCM_STRING_COERCE_0TERMINATION_X(s_attr);
  attr=SCM_STRING_CHARS(s_attr);
  SCMLDAP_CLEAR_ERROR(ld);
  vals=ldap_get_values_len(ld, entry, attr);
  if (vals==NULL) {
    SCMLDAP_SESSION_ERROR(ld);
    return SCM_EOL;
  }
  ct=ldap_count_values_len(vals);
  if (ct==-1) SCMLDAP_ERROR(-1);
  s_vals=SCM_EOL;
  while (ct>0) {
    --ct;
    s_vals=scm_cons(scm_mem2string(vals[ct]->bv_val, vals[ct]->bv_len),
                    s_vals);
  }
  ldap_value_free_len(vals);
  return s_vals;
}
#undef FUNC_NAME

/* initialization */

void scmldap__load__init(void) {
#ifndef SCM_MAGIC_SNARFER
#include "libguile-ldap.guile-snarf"
#endif
}

/*
ldap_first_message
ldap_next_message
ldap_first_reference
ldap_next_reference
ldap_compare
ldap_parse_result

lber_decode
lber_encode
lber_memory
lber_types
ldap_count_entries
ldap_msgid
ldap_msgtype
ldap_result2error

ber_bvecfree
ber_bvfree
ldap_control_free
ldap_controls_free
*/
